from __future__ import annotations

import types

import ulid

EVENT_TYPES = {
    "pytest_runtest_logstart": "start_test",
    "pytest_runtest_logfinish": "end_test",
}


def pytest_generated_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore pytest generated code

    When using the `-k` or `-m` command line argument, pytest compiles
    code with a custom filename:

        <pytest match expression>

    https://github.com/pytest-dev/pytest/blob/9454fc38d3636b79ee657d6cacf7477eb8acee52/src/_pytest/mark/expression.py#L208

    Since this is library code, we want to filter it out.
    """
    return frame.f_code.co_filename == "<pytest match expression>"


def call_pytest(frame, event, arg, context):
    return event == "call"  # pragma: no cover


def build_context(config):
    return {"frame_ids": {}}


def process_pytest(frame, event, arg, context):  # pragma: no cover
    frame_ids = context["frame_ids"]
    co_name = frame.f_code.co_name
    location = frame.f_locals["location"]
    if co_name == "pytest_runtest_logstart":
        frame_id = f"frm_{ulid.new()}"
        frame_ids[id(location)] = frame_id
    else:
        frame_id = frame_ids[id(location)]

    filename, lineno, test = location
    test_class, _sep, test_name = test.rpartition(".")
    return {
        "frame_id": frame_id,
        "type": EVENT_TYPES[co_name],
        "test_name": test_name,
        "test_class": test_class if test_class else None,
    }


pytest = {
    "co_names": tuple(EVENT_TYPES),
    "path_fragment": "kolo/pytest_plugin.py",
    "call": call_pytest,
    "call_type": "",
    "return_type": "",
    "process": process_pytest,
    "build_context": build_context,
}



from __future__ import annotations

import time
import types
from contextlib import suppress

from ..serialize import get_content, get_request_body


def process_django(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    if event == "call":
        context["timestamp"] = time.time()
        request = frame.f_locals["request"]
        return {
            "body": get_request_body(request),
            "headers": dict(request.headers),
            "method": request.method,
            "path_info": request.path_info,
            # dict(request.POST) can give confusing results due to MultivalueDict
            "post_data": dict(request.POST),
            "query_params": dict(request.GET),
            "scheme": request.scheme,
            # Override the timestamp in PluginProcessor because consistency with
            # ms_duration is more important.
            "timestamp": context["timestamp"],
        }
    elif event == "return":  # pragma: no branch
        timestamp = time.time()
        duration = timestamp - context["timestamp"]
        ms_duration = round(duration * 1000, 2)

        request = frame.f_locals["request"]
        match = request.resolver_match
        if match:  # match is None if this is a 404
            # TODO(later): We should record the urlconf used, if it's not ROOT_URLCONF
            # TODO(later): We should report on whether it was a `path` or `re_path` if we can;
            #  we might have to look in `tried` or go up a frame to find the pattern object?
            url_pattern = {
                "namespace": match.namespace,
                "route": match.route,
                "url_name": match.url_name,
                "view_qualname": match._func_path,
            }
        else:
            url_pattern = None

        response = frame.f_locals["response"]
        return {
            "ms_duration": ms_duration,
            "status_code": response.status_code,
            "content": get_content(response),
            "headers": dict(response.items()),
            "url_pattern": url_pattern,
            # Override the timestamp in PluginProcessor because consistency with
            # ms_duration is more important.
            "timestamp": timestamp,
        }


def process_django_template(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    template_context = frame.f_locals["context"]
    with suppress(AttributeError):
        template_context = template_context.flatten()

    return {
        "context": template_context,
        "template": frame.f_locals["self"].template.name,
    }


django = {
    "co_names": ("get_response",),
    "path_fragment": "/kolo/middleware.py",
    "call_type": "django_request",
    "return_type": "django_response",
    "process": process_django,
}

django_template = {
    "co_names": ("render",),
    "path_fragment": "django/template/backends/django.py",
    "call_type": "django_template_start",
    "return_type": "django_template_end",
    "process": process_django_template,
}

django_setup = {
    "co_names": ("setup",),
    "path_fragment": "django/__init__.py",
    "call_type": "django_setup_start",
    "return_type": "django_setup_end",
    "process": None,
}

django_checks = {
    "co_names": ("run_checks",),
    "path_fragment": "django/core/checks/registry.py",
    "call_type": "django_checks_start",
    "return_type": "django_checks_end",
    "process": None,
}

django_test_db = {
    "co_names": ("create_test_db",),
    "path_fragment": "django/db/backends/base/creation.py",
    "call_type": "django_create_test_db_start",
    "return_type": "django_create_test_db_end",
    "process": None,
}



from __future__ import annotations

import logging
import mimetypes
import os
import sqlite3
import sys
import threading
from itertools import chain
from typing import Awaitable, Callable

import httpx
from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.conf import settings
from django.http import (
    FileResponse,
    HttpRequest,
    HttpResponse,
    HttpResponseNotFound,
    JsonResponse,
)

from .config import load_config
from .db import list_traces_from_db, setup_db
from .profiler import KoloProfiler
from .serialize import monkeypatch_queryset_repr
from .upload import upload_to_dashboard
from .web.home import kolo_web_home

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]
DjangoAsyncView = Callable[[HttpRequest], Awaitable[HttpResponse]]


class KoloMiddleware:
    sync_capable = True
    async_capable = True

    def __init__(self, get_response: DjangoView | DjangoAsyncView) -> None:
        self._is_coroutine = iscoroutinefunction(get_response)
        if self._is_coroutine:
            markcoroutinefunction(self)
        self._get_response = get_response
        self.config = load_config()
        self.upload_token = self.get_upload_token()
        self.enabled = self.should_enable()
        if self.enabled:
            self.db_path = setup_db()

            # TODO: Put the full URL here not just the /_kolo/ path
            if self.config.get("web_experience", False):
                print("\nView recent requests at /_kolo/")

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if self._is_coroutine:
            # TODO: Handle _kolo/ for asgi???
            get_response = self.aget_response
        else:
            if request.path.startswith("/_kolo"):
                return kolo_web_router(request)

            get_response = self.get_response  # type: ignore

        # WARNING: Because Django's runserver uses threading, we need
        # to be careful about thread safety here.
        if not self.enabled or self.check_for_third_party_profiler():
            return get_response(request)  # type: ignore

        filter_config = self.config.get("filters", {})
        ignore_request_paths = filter_config.get("ignore_request_paths", [])
        for path in ignore_request_paths:
            if path in request.path:
                return get_response(request)  # type: ignore

        # Don't store the KoloProfiler on self to avoid threadsafety
        # bugs. If a different thread gets this instance of KoloProfiler
        # at the wrong time, we lose the original profiler's trace.
        profiler = KoloProfiler(
            self.db_path, config=self.config, source="kolo.middleware.KoloMiddleware"
        )

        monkeypatch_queryset_repr()
        if self._is_coroutine:
            return self.aprofile_response(request, profiler)
        else:
            return self.profile_response(request, profiler)

    def profile_response(self, request, profiler):
        with profiler:
            response = self.get_response(request)
        self.save_trace(profiler)
        return response

    async def aprofile_response(self, request, profiler):
        with profiler:
            response = await self.aget_response(request)
        self.save_trace(profiler)
        return response

    async def aget_response(self, request: HttpRequest) -> HttpResponse:
        response = await self._get_response(request)  # type: ignore
        return response

    def get_response(self, request: HttpRequest) -> HttpResponse:
        response = self._get_response(request)
        return response  # type: ignore

    def save_trace(self, profiler):
        if settings.DEBUG is False and self.upload_token:

            def upload():
                trace = profiler.build_trace()
                try:
                    response = upload_to_dashboard(trace, self.upload_token)
                    response.raise_for_status()
                except httpx.HTTPError:
                    logger.exception("Failed to upload trace to Kolo dashboard.")

            name = "kolo-upload_to_dashboard"
            threading.Thread(target=upload, name=name).start()
        else:
            name = "kolo-save_request_in_db"
            threading.Thread(target=profiler.save_request_in_db, name=name).start()

    def check_for_third_party_profiler(self) -> bool:
        profiler = sys.getprofile()
        if profiler:
            logger.warning("Profiler %s is active, disabling KoloMiddleware", profiler)
            return True
        if self.config.get("threading", False):
            try:
                profiler = threading.getprofile()  # type: ignore[attr-defined]
            except AttributeError:
                profiler = threading._profile_hook
            if profiler:  # pragma: no branch
                logger.warning(
                    "Profiler %s is active, disabling KoloMiddleware", profiler
                )
                return True
        return False

    def should_enable(self) -> bool:
        if settings.DEBUG is False and self.upload_token is None:
            logger.debug("DEBUG mode is off, disabling KoloMiddleware")
            return False

        if os.environ.get("KOLO_DISABLE", "false").lower() not in ["false", "0"]:
            logger.debug("KOLO_DISABLE is set, disabling KoloMiddleware")
            return False

        if self.check_for_third_party_profiler():
            return False

        return True

    def get_upload_token(self):
        if not self.config.get("production_beta", False):
            return None

        upload_token = os.environ.get("KOLO_API_TOKEN", None)
        if upload_token is None:
            logging.warning(
                "Kolo production beta is enabled, but `KOLO_API_TOKEN` environment variable is not set."
            )
            return None

        if upload_token.startswith("kolo_prod_"):
            return upload_token

        logging.warning("`KOLO_API_TOKEN` is invalid.")
        return None


def kolo_web_router(request: HttpRequest) -> HttpResponse:
    path = request.path
    if path == "/_kolo/" or path == "/_kolo" or path.startswith("/_kolo/traces/"):
        return kolo_web_home(request)
    elif path.startswith("/_kolo/static/"):
        static_dir = os.path.join(os.path.dirname(__file__), "web", "static")
        file_path = os.path.join(static_dir, request.path[len("/_kolo/static/") :])

        if os.path.exists(file_path):
            mime_type, encoding = mimetypes.guess_type(file_path)
            return FileResponse(  # type: ignore # seems iffy and difficult to fix with little benefit
                open(file_path, "rb"),
                content_type=mime_type or "application/octet-stream",
            )
        else:
            return HttpResponseNotFound("File not found")
    elif path.startswith("/_kolo/api/generate-test/"):
        return kolo_web_api_generate_test(request)
    elif path.startswith("/_kolo/api/traces/"):
        return kolo_web_api_get_trace(request)
    elif path.startswith("/_kolo/api/latest-traces/"):
        return kolo_web_api_latest_traces(request)
    else:
        return HttpResponseNotFound("Kolo Web: Not Found")


def kolo_web_api_generate_test(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/generate-test/", "").replace("/", "")

    from .generate_tests import generate_from_trace_ids

    test_code = generate_from_trace_ids(
        trace_id, test_class="MyTestCase", test_name="test_my_view"
    )

    return JsonResponse({"test_code": test_code})


def kolo_web_api_get_trace(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/traces/", "").replace("/", "")

    from .db import load_trace_from_db

    db_path = setup_db()

    msgpack_data, json_data = load_trace_from_db(db_path, trace_id)
    if json_data:
        return HttpResponse(json_data, content_type="application/json")
    else:
        return HttpResponse(msgpack_data, content_type="application/msgpack")


def kolo_web_api_latest_traces(request: HttpRequest) -> HttpResponse:
    from .db import db_connection

    db_path = setup_db()

    with db_connection(db_path) as connection:
        needs_reversed_order = False
        reached_top = False

        if "anchor" in request.GET and "showNext" in request.GET:
            # this is a pagination request

            anchor = request.GET["anchor"]
            show_next = int(request.GET["showNext"])

            limit = abs(show_next)

            # Positive show_next value means we're going back in time (loading _more_ traces),
            # negative value is for going forward in time (loading traces the user has previously seen before going back).

            if show_next > 0:
                # going back in time, trying to access older traces than the anchor

                query = "SELECT id FROM traces WHERE id < ? ORDER BY id desc LIMIT ?"
                cursor = connection.execute(query, (anchor, limit))
                rows = cursor.fetchall()
            else:
                # going forward in time, trying to access newer traces than the anchor

                query = "SELECT id FROM traces WHERE id > ? ORDER BY id LIMIT ?"
                needs_reversed_order = True
                # In order to get 10 newer traces, they need to be sorted in ascending order.
                # They have to be reversed later because the endpoint should always return traces from newest to oldest.

                cursor = connection.execute(query, (anchor, limit))
                rows = cursor.fetchall()

                if len(rows) < abs(limit):
                    # If there are less than 10 newer traces, we need to fetch some older traces to fill up the response.
                    cursor = connection.execute(
                        "SELECT id FROM traces ORDER BY id desc LIMIT ?", (abs(limit),)
                    )
                    rows = cursor.fetchall()

                    reached_top = True
                    needs_reversed_order = False
        else:
            # not a pagination request, we just want N latest traces

            limit = int(request.GET.get("showNext", 10))
            cursor = connection.execute(
                "SELECT id FROM traces ORDER BY id desc LIMIT ?", (limit,)
            )

            rows = cursor.fetchall()
            reached_top = True

        traces = list(chain.from_iterable(rows))

        if needs_reversed_order:
            traces = traces[::-1]

    return JsonResponse({"traces": traces, "isTop": reached_top})



from __future__ import annotations

import logging
from typing import Awaitable, Callable

from django.http import HttpRequest, HttpResponse

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]
DjangoAsyncView = Callable[[HttpRequest], Awaitable[HttpResponse]]


def kolo_web_home(request: HttpRequest) -> HttpResponse:
    html = f"""<!DOCTYPE html>
  <head>
      <meta charset="UTF-8">
      <meta name="viewport" content="width=device-width, initial-scale=1.0">
      <title>Kolo</title>

      <link rel="icon" href="/_kolo/favicon.ico">

      <!-- For light mode -->
      <link rel="icon" sizes="16x16" href="/_kolo/static/favicons/favicon-dark-16x16.png" media="(prefers-color-scheme: light)">
      <link rel="icon" sizes="32x32" href="/_kolo/static/favicons/favicon-dark-32x32.png" media="(prefers-color-scheme: light)">

      <!-- For dark mode -->
      <link rel="icon" sizes="16x16" href="/_kolo/static/favicons/favicon-light-16x16.png" media="(prefers-color-scheme: dark)">
      <link rel="icon" sizes="32x32" href="/_kolo/static/favicons/favicon-light-32x32.png" media="(prefers-color-scheme: dark)">

      <link rel="stylesheet" href="/_kolo/static/main.css" />
      <script type="text/javascript">
      window.kolo = {{
        "options": {{
            "mountPoint": "/_kolo/",
            "baseUrl": new URL('/_kolo/', window.location),
            "island": false,
            "local": true
        }}
      }}
    </script>
    </head>
  <body>
    <div id="root"></div>
    <script type="text/javascript" src="/_kolo/static/main.js"></script>
  </body>
"""

    return HttpResponse(html)
